import pytest

from deltalake import DeltaTable


@pytest.mark.no_pyarrow
def test_import_error_reading_pyarrow(existing_sample_table: DeltaTable):
    with pytest.raises(ImportError, match="pyarrow"):
        existing_sample_table.to_pyarrow_dataset()

    with pytest.raises(ImportError, match="pyarrow"):
        existing_sample_table.to_pyarrow_table()

    with pytest.raises(ImportError, match="pyarrow"):
        existing_sample_table.to_pandas()


@pytest.mark.no_pyarrow
def test_import_error_storage_handler_from_table(existing_sample_table: DeltaTable):
    from deltalake.fs import DeltaStorageHandler

    with pytest.raises(ImportError, match="pyarrow"):
        DeltaStorageHandler.from_table(existing_sample_table)


@pytest.mark.no_pyarrow
def test_import_error_storage_handler_init():
    from deltalake.fs import DeltaStorageHandler

    with pytest.raises(ImportError, match="pyarrow"):
        DeltaStorageHandler(table_uri="random")



from deltalake._internal import (
    CommitFailedError,
    DeltaError,
    DeltaProtocolError,
    SchemaMismatchError,
    TableNotFoundError,
)

__all__ = [
    "CommitFailedError",
    "DeltaError",
    "DeltaProtocolError",
    "SchemaMismatchError",
    "TableNotFoundError",
]



from __future__ import annotations

from collections.abc import Mapping
from enum import Enum
from typing import (
    TYPE_CHECKING,
    Any,
    Literal,
    Optional,
    Union,
)

from arro3.core import DataType as ArrowDataType
from arro3.core import Field as ArrowField
from arro3.core import RecordBatch, RecordBatchReader
from arro3.core import Schema as ArrowSchema
from arro3.core.types import ArrowSchemaExportable

if TYPE_CHECKING:
    import pyarrow.fs as fs

    from deltalake.transaction import (
        AddAction,
        CommitProperties,
        PostCommitHookProperties,
    )
    from deltalake.writer.properties import (
        WriterProperties,
    )
__version__: str

class TableFeatures(Enum):
    # Mapping of one column to another
    ColumnMapping = "ColumnMapping"
    # Deletion vectors for merge, update, delete
    DeletionVectors = "DeletionVectors"
    # timestamps without timezone support
    TimestampWithoutTimezone = "TimestampWithoutTimezone"
    # version 2 of checkpointing
    V2Checkpoint = "V2Checkpoint"
    # Append Only Tables
    AppendOnly = "AppendOnly"
    # Table invariants
    Invariants = "Invariants"
    # Check constraints on columns
    CheckConstraints = "CheckConstraints"
    # CDF on a table
    ChangeDataFeed = "ChangeDataFeed"
    # Columns with generated values
    GeneratedColumns = "GeneratedColumns"
    # ID Columns
    IdentityColumns = "IdentityColumns"
    # Row tracking on tables
    RowTracking = "RowTracking"
    # domain specific metadata
    DomainMetadata = "DomainMetadata"
    # Iceberg compatibility support
    IcebergCompatV1 = "IcebergCompatV1"

class RawDeltaTableMetaData:
    id: int
    name: str
    description: str
    partition_columns: list[str]
    created_time: int
    configuration: dict[str, str]

class RawDeltaTable:
    schema: Any

    def __init__(
        self,
        table_uri: str,
        version: int | None,
        storage_options: dict[str, str] | None,
        without_files: bool,
        log_buffer_size: int | None,
    ) -> None: ...
    @staticmethod
    def get_table_uri_from_data_catalog(
        data_catalog: str,
        database_name: str,
        table_name: str,
        data_catalog_id: str | None = None,
        catalog_options: dict[str, str] | None = None,
    ) -> str: ...
    @staticmethod
    def is_deltatable(
        table_uri: str, storage_options: dict[str, str] | None
    ) -> bool: ...
    def table_uri(self) -> str: ...
    def version(self) -> int: ...
    def has_files(self) -> bool: ...
    def get_add_file_sizes(self) -> dict[str, int]: ...
    def get_latest_version(self) -> int: ...
    def metadata(self) -> RawDeltaTableMetaData: ...
    def protocol_versions(
        self,
    ) -> tuple[int, int, Optional[list[str]], Optional[list[str]]]: ...
    def table_config(self) -> tuple[bool, int]: ...
    def load_version(self, version: int) -> None: ...
    def load_with_datetime(self, ds: str) -> None: ...
    def files(self, partition_filters: PartitionFilterType | None) -> list[str]: ...
    def file_uris(self, partition_filters: PartitionFilterType | None) -> list[str]: ...
    def vacuum(
        self,
        dry_run: bool,
        retention_hours: int | None,
        enforce_retention_duration: bool,
        commit_properties: CommitProperties | None,
        post_commithook_properties: PostCommitHookProperties | None,
        full: bool,
    ) -> list[str]: ...
    def compact_optimize(
        self,
        partition_filters: PartitionFilterType | None,
        target_size: int | None,
        max_concurrent_tasks: int | None,
        min_commit_interval: int | None,
        writer_properties: WriterProperties | None,
        commit_properties: CommitProperties | None,
        post_commithook_properties: PostCommitHookProperties | None,
    ) -> str: ...
    def z_order_optimize(
        self,
        z_order_columns: list[str],
        partition_filters: PartitionFilterType | None,
        target_size: int | None,
        max_concurrent_tasks: int | None,
        max_spill_size: int | None,
        min_commit_interval: int | None,
        writer_properties: WriterProperties | None,
        commit_properties: CommitProperties | None,
        post_commithook_properties: PostCommitHookProperties | None,
    ) -> str: ...
    def add_columns(
        self,
        fields: list[Field],
        commit_properties: CommitProperties | None,
        post_commithook_properties: PostCommitHookProperties | None,
    ) -> None: ...
    def add_feature(
        self,
        feature: list[TableFeatures],
        allow_protocol_versions_increase: bool,
        commit_properties: CommitProperties | None,
        post_commithook_properties: PostCommitHookProperties | None,
    ) -> None: ...
    def add_constraints(
        self,
        constraints: dict[str, str],
        commit_properties: CommitProperties | None,
        post_commithook_properties: PostCommitHookProperties | None,
    ) -> None: ...
    def drop_constraints(
        self,
        name: str,
        raise_if_not_exists: bool,
        commit_properties: CommitProperties | None,
        post_commithook_properties: PostCommitHookProperties | None,
    ) -> None: ...
    def set_table_properties(
        self,
        properties: dict[str, str],
        raise_if_not_exists: bool,
        commit_properties: CommitProperties | None,
    ) -> None: ...
    def set_table_name(
        self,
        name: str,
        commit_properties: CommitProperties | None = None,
    ) -> None: ...
    def set_table_description(
        self,
        description: str,
        commit_properties: CommitProperties | None = None,
    ) -> None: ...
    def restore(
        self,
        target: Any | None,
        ignore_missing_files: bool,
        protocol_downgrade_allowed: bool,
        commit_properties: CommitProperties | None,
    ) -> str: ...
    def history(self, limit: int | None) -> list[str]: ...
    def update_incremental(self) -> None: ...
    def dataset_partitions(
        self,
        schema: ArrowSchemaExportable,
        partition_filters: FilterConjunctionType | None,
    ) -> list[Any]: ...
    def create_checkpoint(self) -> None: ...
    def get_add_actions(self, flatten: bool) -> RecordBatch: ...
    def delete(
        self,
        predicate: str | None,
        writer_properties: WriterProperties | None,
        commit_properties: CommitProperties | None,
        post_commithook_properties: PostCommitHookProperties | None,
    ) -> str: ...
    def repair(
        self,
        dry_run: bool,
        commit_properties: CommitProperties | None,
        post_commithook_properties: PostCommitHookProperties | None,
    ) -> str: ...
    def update(
        self,
        updates: dict[str, str],
        predicate: str | None,
        writer_properties: WriterProperties | None,
        safe_cast: bool,
        commit_properties: CommitProperties | None,
        post_commithook_properties: PostCommitHookProperties | None,
    ) -> str: ...
    def create_merge_builder(
        self,
        source: RecordBatchReader,
        batch_schema: ArrowSchema,
        predicate: str,
        source_alias: str | None,
        target_alias: str | None,
        merge_schema: bool,
        writer_properties: WriterProperties | None,
        commit_properties: CommitProperties | None,
        post_commithook_properties: PostCommitHookProperties | None,
        safe_cast: bool,
        streamed_exec: bool,
    ) -> PyMergeBuilder: ...
    def merge_execute(self, merge_builder: PyMergeBuilder) -> str: ...
    def get_active_partitions(
        self, partitions_filters: FilterType | None = None
    ) -> Any: ...
    def create_write_transaction(
        self,
        add_actions: list[AddAction],
        mode: str,
        partition_by: list[str],
        schema: Schema,
        partitions_filters: FilterType | None,
        commit_properties: CommitProperties | None,
        post_commithook_properties: PostCommitHookProperties | None,
    ) -> None: ...
    def cleanup_metadata(self) -> None: ...
    def load_cdf(
        self,
        columns: list[str] | None = None,
        predicate: str | None = None,
        starting_version: int = 0,
        ending_version: int | None = None,
        starting_timestamp: str | None = None,
        ending_timestamp: str | None = None,
        allow_out_of_range: bool = False,
    ) -> RecordBatchReader: ...
    def transaction_version(self, app_id: str) -> int | None: ...
    def set_column_metadata(
        self,
        column: str,
        metadata: dict[str, str],
        commit_properties: CommitProperties | None,
        post_commithook_properties: PostCommitHookProperties | None,
    ) -> None: ...
    def __datafusion_table_provider__(self) -> Any: ...
    def write(
        self,
        data: RecordBatchReader,
        batch_schema: ArrowSchema,
        partition_by: list[str] | None,
        mode: str,
        schema_mode: str | None,
        predicate: str | None,
        target_file_size: int | None,
        name: str | None,
        description: str | None,
        configuration: Mapping[str, str | None] | None,
        writer_properties: WriterProperties | None,
        commit_properties: CommitProperties | None,
        post_commithook_properties: PostCommitHookProperties | None,
    ) -> None: ...

def rust_core_version() -> str: ...
def create_table_with_add_actions(
    table_uri: str,
    schema: Schema,
    add_actions: list[AddAction],
    mode: Literal["error", "append", "overwrite", "ignore"],
    partition_by: list[str],
    name: str | None,
    description: str | None,
    configuration: Mapping[str, str | None] | None,
    storage_options: dict[str, str] | None,
    commit_properties: CommitProperties | None,
    post_commithook_properties: PostCommitHookProperties | None,
) -> None: ...
def write_to_deltalake(
    table_uri: str,
    data: RecordBatchReader,
    batch_schema: ArrowSchema,
    partition_by: list[str] | None,
    mode: str,
    schema_mode: str | None,
    predicate: str | None,
    target_file_size: int | None,
    name: str | None,
    description: str | None,
    configuration: Mapping[str, str | None] | None,
    storage_options: dict[str, str] | None,
    writer_properties: WriterProperties | None,
    commit_properties: CommitProperties | None,
    post_commithook_properties: PostCommitHookProperties | None,
) -> None: ...
def convert_to_deltalake(
    uri: str,
    partition_schema: Schema | None,
    partition_strategy: Literal["hive"] | None,
    name: str | None,
    description: str | None,
    configuration: Mapping[str, str | None] | None,
    storage_options: dict[str, str] | None,
    commit_properties: CommitProperties | None,
    post_commithook_properties: PostCommitHookProperties | None,
) -> None: ...
def create_deltalake(
    table_uri: str,
    schema: Schema,
    partition_by: list[str],
    mode: str,
    raise_if_key_not_exists: bool,
    name: str | None,
    description: str | None,
    configuration: Mapping[str, str | None] | None,
    storage_options: dict[str, str] | None,
    commit_properties: CommitProperties | None,
    post_commithook_properties: PostCommitHookProperties | None,
) -> None: ...
def get_num_idx_cols_and_stats_columns(
    table: RawDeltaTable | None, configuration: Mapping[str, str | None] | None
) -> tuple[int, list[str] | None]: ...

class PyMergeBuilder:
    source_alias: str
    target_alias: str
    merge_schema: bool
    arrow_schema: ArrowSchema

    def when_matched_update(
        self, updates: dict[str, str], predicate: str | None
    ) -> None: ...
    def when_matched_delete(self, predicate: str | None) -> None: ...
    def when_not_matched_insert(
        self, updates: dict[str, str], predicate: str | None
    ) -> None: ...
    def when_not_matched_by_source_update(
        self, updates: dict[str, str], predicate: str | None
    ) -> None: ...
    def when_not_matched_by_source_delete(
        self,
        predicate: str | None,
    ) -> None: ...

# Can't implement inheritance (see note in src/schema.rs), so this is next
# best thing.
DataType = Union["PrimitiveType", "MapType", "StructType", "ArrayType"]

class PrimitiveType:
    """A primitive datatype, such as a string or number.

    Can be initialized with a string value:

    ```
    PrimitiveType("integer")
    ```

    Valid primitive data types include:

     * "string",
     * "long",
     * "integer",
     * "short",
     * "byte",
     * "float",
     * "double",
     * "boolean",
     * "binary",
     * "date",
     * "timestamp",
     * "timestamp_ntz",
     * "decimal(<precision>, <scale>)" Max: decimal(38,38)

    Args:
        data_type: string representation of the data type
    """

    def __init__(self, data_type: str) -> None: ...
    type: str
    """ The inner type
    """

    def to_json(self) -> str: ...
    @staticmethod
    def from_json(json: str) -> PrimitiveType:
        """Create a PrimitiveType from a JSON string

        The JSON representation for a primitive type is just a quoted string: `PrimitiveType.from_json('"integer"')`

        Args:
            json: a JSON string

        Returns:
            a PrimitiveType type
        """
    def to_arrow(self) -> ArrowDataType:
        """Get the equivalent arro3 DataType (arro3.core.DataType)"""

    @staticmethod
    def from_arrow(type: ArrowSchemaExportable) -> PrimitiveType:
        """Create a PrimitiveType from an `ArrowSchemaExportable` datatype

        Will raise `TypeError` if the arrow type is not a primitive type.

        Args:
            type: an object that is `ArrowSchemaExportable`

        Returns:
            a PrimitiveType
        """
    def __arrow_c_schema__(self) -> object:
        """
        An implementation of the [Arrow PyCapsule
        Interface](https://arrow.apache.org/docs/format/CDataInterface/PyCapsuleInterface.html).
        This dunder method should not be called directly, but enables zero-copy data
        transfer to other Python libraries that understand Arrow memory.

        For example, you can call [`pyarrow.schema()`][pyarrow.schema] to convert this
        array into a pyarrow schema, without copying memory.
        """

class ArrayType:
    """An Array (List) DataType

    Example:
        Can either pass the element type explicitly or can pass a string
        if it is a primitive type:
        ```python
        ArrayType(PrimitiveType("integer"))
        # Returns ArrayType(PrimitiveType("integer"), contains_null=True)

        ArrayType("integer", contains_null=False)
        # Returns ArrayType(PrimitiveType("integer"), contains_null=False)
        ```
    """

    def __init__(
        self, element_type: DataType, *, contains_null: bool = True
    ) -> None: ...
    type: Literal["array"]
    """ The string "array"
    """

    element_type: DataType
    """ The type of the element, of type:
        Union[
            [PrimitiveType][deltalake.schema.PrimitiveType],
            [ArrayType][deltalake.schema.ArrayType],
            [MapType][deltalake.schema.MapType],
            [StructType][deltalake.schema.StructType]
        ]
    """

    contains_null: bool
    """ Whether the arrays may contain null values
    """

    def to_json(self) -> str:
        """Get the JSON string representation of the type."""
    @staticmethod
    def from_json(json: str) -> ArrayType:
        """Create an ArrayType from a JSON string

        Args:
            json: a JSON string

        Returns:
            an ArrayType

        Example:
            The JSON representation for an array type is an object with `type` (set to
            `"array"`), `elementType`, and `containsNull`.
            ```python
            ArrayType.from_json(
                '''{
                    "type": "array",
                    "elementType": "integer",
                    "containsNull": false
                }'''
            )
            # Returns ArrayType(PrimitiveType("integer"), contains_null=False)
            ```
        """
    def to_arrow(
        self,
    ) -> ArrowDataType:
        """Get the equivalent arro3 type."""
    @staticmethod
    def from_arrow(type: ArrowSchemaExportable) -> ArrayType:
        """Create an ArrayType from an `ArrowSchemaExportable` datatype.

        Will raise `TypeError` if a different arrow DataType is provided.

        Args:
            type: an object that is `ArrowSchemaExportable`

        Returns:
            an ArrayType
        """
    def __arrow_c_schema__(self) -> object:
        """
        An implementation of the [Arrow PyCapsule
        Interface](https://arrow.apache.org/docs/format/CDataInterface/PyCapsuleInterface.html).
        This dunder method should not be called directly, but enables zero-copy data
        transfer to other Python libraries that understand Arrow memory.

        For example, you can call [`pyarrow.schema()`][pyarrow.schema] to convert this
        array into a pyarrow schema, without copying memory.
        """

class MapType:
    """A map data type

    `key_type` and `value_type` should be [PrimitiveType][deltalake.schema.PrimitiveType], [ArrayType][deltalake.schema.ArrayType],
    or [StructType][deltalake.schema.StructType]. A string can also be passed, which will be
    parsed as a primitive type:

    Example:
        ```python
        MapType(PrimitiveType("integer"), PrimitiveType("string"))
        # Returns MapType(PrimitiveType("integer"), PrimitiveType("string"), value_contains_null=True)

        MapType("integer", "string", value_contains_null=False)
        # Returns MapType(PrimitiveType("integer"), PrimitiveType("string"), value_contains_null=False)
        ```
    """

    def __init__(
        self,
        key_type: DataType,
        value_type: DataType,
        *,
        value_contains_null: bool = True,
    ) -> None: ...
    type: Literal["map"]
    key_type: DataType
    """ The type of the keys, of type:
        Union[
            [PrimitiveType][deltalake.schema.PrimitiveType],
            [ArrayType][deltalake.schema.ArrayType],
            [MapType][deltalake.schema.MapType],
            [StructType][deltalake.schema.StructType]
        ]
    """

    value_type: DataType
    """The type of the values, of type:
        Union[
            [PrimitiveType][deltalake.schema.PrimitiveType],
            [ArrayType][deltalake.schema.ArrayType],
            [MapType][deltalake.schema.MapType],
            [StructType][deltalake.schema.StructType]
        ]
    """

    value_contains_null: bool
    """ Whether the values in a map may be null
    """

    def to_json(self) -> str:
        """Get JSON string representation of map type.

        Returns:
            a JSON string
        """
    @staticmethod
    def from_json(json: str) -> MapType:
        """Create a MapType from a JSON string

        Args:
            json: a JSON string

        Returns:
            an ArrayType

        Example:
            The JSON representation for a map type is an object with `type` (set to `map`),
            `keyType`, `valueType`, and `valueContainsNull`:

            ```python
            MapType.from_json(
                '''{
                    "type": "map",
                    "keyType": "integer",
                    "valueType": "string",
                    "valueContainsNull": true
                }'''
            )
            # Returns MapType(PrimitiveType("integer"), PrimitiveType("string"), value_contains_null=True)
            ```
        """
    def to_arrow(self) -> ArrowDataType:
        """Get the equivalent arro3 data type."""

    @staticmethod
    def from_arrow(type: ArrowSchemaExportable) -> MapType:
        """Create a MapType from an `ArrowSchemaExportable` datatype

        Will raise `TypeError` if passed a different type.

        Args:
            type: an object that is `ArrowSchemaExportable`

        Returns:
            a MapType
        """
    def __arrow_c_schema__(self) -> object:
        """
        An implementation of the [Arrow PyCapsule
        Interface](https://arrow.apache.org/docs/format/CDataInterface/PyCapsuleInterface.html).
        This dunder method should not be called directly, but enables zero-copy data
        transfer to other Python libraries that understand Arrow memory.

        For example, you can call [`pyarrow.schema()`][pyarrow.schema] to convert this
        array into a pyarrow schema, without copying memory.
        """

class Field:
    """A field in a Delta StructType or Schema

    Example:
        Can create with just a name and a type:
        ```python
        Field("my_int_col", "integer")
        # Returns Field("my_int_col", PrimitiveType("integer"), nullable=True, metadata=None)
        ```

        Can also attach metadata to the field. Metadata should be a dictionary with
        string keys and JSON-serializable values (str, list, int, float, dict):

        ```python
        Field("my_col", "integer", metadata={"custom_metadata": {"test": 2}})
        # Returns Field("my_col", PrimitiveType("integer"), nullable=True, metadata={"custom_metadata": {"test": 2}})
        ```
    """

    def __init__(
        self,
        name: str,
        type: DataType,
        *,
        nullable: bool = True,
        metadata: dict[str, Any] | None = None,
    ) -> None: ...
    name: str
    """ The name of the field
    """

    type: DataType
    """ The type of the field, of type:
        Union[
            [PrimitiveType][deltalake.schema.PrimitiveType],
            [ArrayType][deltalake.schema.ArrayType],
            [MapType][deltalake.schema.MapType],
            [StructType][deltalake.schema.StructType]
        ]
    """

    nullable: bool
    """ Whether there may be null values in the field
    """

    metadata: dict[str, Any]
    """ The metadata of the field
    """

    def to_json(self) -> str:
        """Get the field as JSON string.

        Returns:
            a JSON string

        Example:
            ```python
            Field("col", "integer").to_json()
            # Returns '{"name":"col","type":"integer","nullable":true,"metadata":{}}'
            ```
        """
    @staticmethod
    def from_json(json: str) -> Field:
        """Create a Field from a JSON string.

        Args:
            json: the JSON string.

        Returns:
            Field

        Example:
            ```
            Field.from_json('''{
                    "name": "col",
                    "type": "integer",
                    "nullable": true,
                    "metadata": {}
                }'''
            )
            # Returns Field(col, PrimitiveType("integer"), nullable=True)
            ```
        """
    def to_arrow(self) -> ArrowField:
        """Convert to an equivalent arro3 field
        Note: This currently doesn't preserve field metadata.

        Returns:
            a arro3 Field
        """
    @staticmethod
    def from_arrow(field: ArrowSchemaExportable) -> Field:
        """Create a Field from an object with an `ArrowSchemaExportable` field

        Note: This currently doesn't preserve field metadata.

        Args:
            field: a Field object that is `ArrowSchemaExportable`

        Returns:
            a Field
        """
    def __arrow_c_schema__(self) -> object:
        """
        An implementation of the [Arrow PyCapsule
        Interface](https://arrow.apache.org/docs/format/CDataInterface/PyCapsuleInterface.html).
        This dunder method should not be called directly, but enables zero-copy data
        transfer to other Python libraries that understand Arrow memory.

        For example, you can call [`pyarrow.schema()`][pyarrow.schema] to convert this
        array into a pyarrow schema, without copying memory.
        """

class StructType:
    """A struct datatype, containing one or more subfields

    Example:
        Create with a list of :class:`Field`:
        ```python
        StructType([Field("x", "integer"), Field("y", "string")])
        # Creates: StructType([Field(x, PrimitiveType("integer"), nullable=True), Field(y, PrimitiveType("string"), nullable=True)])
        ```
    """

    def __init__(self, fields: list[Field]) -> None: ...
    type: Literal["struct"]
    fields: list[Field]
    """ The fields within the struct
    """

    def to_json(self) -> str:
        """Get the JSON representation of the type.

        Returns:
            a JSON string

        Example:
            ```python
            StructType([Field("x", "integer")]).to_json()
            # Returns '{"type":"struct","fields":[{"name":"x","type":"integer","nullable":true,"metadata":{}}]}'
            ```
        """
    @staticmethod
    def from_json(json: str) -> StructType:
        """Create a new StructType from a JSON string.

        Args:
            json: a JSON string

        Returns:
            a StructType

        Example:
            ```python
            StructType.from_json(
                '''{
                    "type": "struct",
                    "fields": [{"name": "x", "type": "integer", "nullable": true, "metadata": {}}]
                }'''
            )
            # Returns StructType([Field(x, PrimitiveType("integer"), nullable=True)])
            ```
        """
    def to_arrow(self) -> ArrowDataType:
        """Get the equivalent arro3 DataType (arro3.core.DataType)"""

    @staticmethod
    def from_arrow(type: ArrowSchemaExportable) -> StructType:
        """Create a new StructType from an `ArrowSchemaExportable` datatype

        Will raise `TypeError` if a different data type is provided.

        Args:
            type: a struct type object that is `ArrowSchemaExportable`

        Returns:
            a StructType
        """
    def __arrow_c_schema__(self) -> object:
        """
        An implementation of the [Arrow PyCapsule
        Interface](https://arrow.apache.org/docs/format/CDataInterface/PyCapsuleInterface.html).
        This dunder method should not be called directly, but enables zero-copy data
        transfer to other Python libraries that understand Arrow memory.

        For example, you can call [`pyarrow.schema()`][pyarrow.schema] to convert this
        array into a pyarrow schema, without copying memory.
        """

class Schema:
    def __init__(self, fields: list[Field]) -> None: ...
    fields: list[Field]

    invariants: list[tuple[str, str]]
    """ The list of invariants on the table. Each invarint is a tuple of strings. The first string is the
        field path and the second is the SQL of the invariant.
    """
    def to_json(self) -> str:
        """Get the JSON string representation of the Schema.

        Returns:
            a JSON string

        Example:
            A schema has the same JSON format as a StructType.
            ```python
            Schema([Field("x", "integer")]).to_json()
            # Returns '{"type":"struct","fields":[{"name":"x","type":"integer","nullable":true,"metadata":{}}]}'
            ```
        """
    @staticmethod
    def from_json(json: str) -> Schema:
        """Create a new Schema from a JSON string.

        Args:
            json: a JSON string

        Example:
            A schema has the same JSON format as a StructType.
            ```python
            Schema.from_json('''{
                "type": "struct",
                "fields": [{"name": "x", "type": "integer", "nullable": true, "metadata": {}}]
                }
            )'''
            # Returns Schema([Field(x, PrimitiveType("integer"), nullable=True)])
            ```
        """
    def to_arrow(self, as_large_types: bool = False) -> ArrowSchema:
        """Return equivalent arro3 schema

        Args:
            as_large_types: get schema with all variable size types (list, binary, string) as large variants (with int64 indices).
                This is for compatibility with systems like Polars that only support the large versions of Arrow types.

        Returns:
            an arro3 Schema
        """

    @staticmethod
    def from_arrow(type: ArrowSchemaExportable) -> Schema:
        """Create a [Schema][deltalake.schema.Schema] from a schema that implements Arrow C Data Interface.

        Will raise `TypeError` if one of the Arrow type is not a primitive type.

        Args:
            type: an object that is `ArrowSchemaExportable`

        Returns:
            a Schema
        """

    def __arrow_c_schema__(self) -> object:
        """
        An implementation of the [Arrow PyCapsule
        Interface](https://arrow.apache.org/docs/format/CDataInterface/PyCapsuleInterface.html).
        This dunder method should not be called directly, but enables zero-copy data
        transfer to other Python libraries that understand Arrow memory.

        For example, you can call [`pyarrow.schema()`][pyarrow.schema] to convert this
        array into a pyarrow schema, without copying memory.
        """

class ObjectInputFile:
    @property
    def closed(self) -> bool: ...
    @property
    def mode(self) -> str: ...
    def isatty(self) -> bool: ...
    def readable(self) -> bool: ...
    def seekable(self) -> bool: ...
    def tell(self) -> int: ...
    def size(self) -> int: ...
    def seek(self, position: int, whence: int = 0) -> int: ...
    def read(self, nbytes: int) -> bytes: ...

class ObjectOutputStream:
    @property
    def closed(self) -> bool: ...
    @property
    def mode(self) -> str: ...
    def isatty(self) -> bool: ...
    def readable(self) -> bool: ...
    def seekable(self) -> bool: ...
    def writable(self) -> bool: ...
    def tell(self) -> int: ...
    def size(self) -> int: ...
    def seek(self, position: int, whence: int = 0) -> int: ...
    def read(self, nbytes: int) -> bytes: ...
    def write(self, data: bytes) -> int: ...

class DeltaFileSystemHandler:
    """Implementation of [pyarrow.fs.FileSystemHandler][pyarrow.fs.FileSystemHandler] for use with [pyarrow.fs.PyFileSystem][pyarrow.fs.PyFileSystem]"""

    def __init__(
        self,
        table_uri: str,
        options: dict[str, str] | None = None,
        known_sizes: dict[str, int] | None = None,
    ) -> None: ...
    @classmethod
    def from_table(
        cls,
        table: RawDeltaTable,
        options: dict[str, str] | None = None,
        known_sizes: dict[str, int] | None = None,
    ) -> DeltaFileSystemHandler: ...
    def get_type_name(self) -> str: ...
    def copy_file(self, src: str, dst: str) -> None:
        """Copy a file.

        If the destination exists and is a directory, an error is returned. Otherwise, it is replaced.
        """
    def create_dir(self, path: str, recursive: bool = True) -> None:
        """Create a directory and subdirectories.

        This function succeeds if the directory already exists.
        """
    def delete_dir(self, path: str) -> None:
        """Delete a directory and its contents, recursively."""
    def delete_file(self, path: str) -> None:
        """Delete a file."""
    def equals(self, other: Any) -> bool: ...
    def delete_dir_contents(
        self, path: str, *, accept_root_dir: bool = False, missing_dir_ok: bool = False
    ) -> None:
        """Delete a directory's contents, recursively.

        Like delete_dir, but doesn't delete the directory itself.
        """
    def delete_root_dir_contents(self) -> None:
        """Delete the root directory contents, recursively."""
    def get_file_info(self, paths: list[str]) -> list[fs.FileInfo]:
        """Get info for the given files.

        A non-existing or unreachable file returns a FileStat object and has a FileType of value NotFound.
        An exception indicates a truly exceptional condition (low-level I/O error, etc.).
        """
    def get_file_info_selector(
        self, base_dir: str, allow_not_found: bool = False, recursive: bool = False
    ) -> list[fs.FileInfo]:
        """Get info for the given files.

        A non-existing or unreachable file returns a FileStat object and has a FileType of value NotFound.
        An exception indicates a truly exceptional condition (low-level I/O error, etc.).
        """
    def move_file(self, src: str, dest: str) -> None:
        """Move / rename a file or directory.

        If the destination exists: - if it is a non-empty directory, an error is returned - otherwise,
        if it has the same type as the source, it is replaced - otherwise, behavior is
        unspecified (implementation-dependent).
        """
    def normalize_path(self, path: str) -> str:
        """Normalize filesystem path."""
    def open_input_file(self, path: str) -> ObjectInputFile:
        """Open an input file for random access reading."""
    def open_output_stream(
        self, path: str, metadata: dict[str, str] | None = None
    ) -> ObjectOutputStream:
        """Open an output stream for sequential writing."""

class PyQueryBuilder:
    def __init__(self) -> None: ...
    def register(self, table_name: str, delta_table: RawDeltaTable) -> None: ...
    def execute(self, sql: str) -> RecordBatchReader: ...

class DeltaError(Exception):
    """The base class for Delta-specific errors."""

    pass

class TableNotFoundError(DeltaError):
    """Raised when a Delta table cannot be loaded from a location."""

    pass

class CommitFailedError(DeltaError):
    """Raised when a commit to a Delta table fails."""

    pass

class DeltaProtocolError(DeltaError):
    """Raised when a violation with the Delta protocol specs occurred."""

    pass

class SchemaMismatchError(DeltaError):
    """Raised when a schema mismatch is detected."""

    pass

FilterLiteralType = tuple[str, str, Any]
FilterConjunctionType = list[FilterLiteralType]
FilterDNFType = list[FilterConjunctionType]
FilterType = FilterConjunctionType | FilterDNFType
PartitionFilterType = list[tuple[str, str, str | list[str]]]

class Transaction:
    app_id: str
    version: int
    last_updated: int | None

    def __init__(
        self, app_id: str, version: int, last_updated: int | None = None
    ) -> None: ...



from deltalake._internal import (
    TableFeatures,
    Transaction,
    __version__,
    rust_core_version,
)
from deltalake.query import QueryBuilder
from deltalake.schema import DataType, Field, Schema
from deltalake.table import DeltaTable, Metadata
from deltalake.transaction import CommitProperties, PostCommitHookProperties
from deltalake.writer import (
    BloomFilterProperties,
    ColumnProperties,
    WriterProperties,
    convert_to_deltalake,
    write_deltalake,
)

__all__ = [
    "BloomFilterProperties",
    "ColumnProperties",
    "CommitProperties",
    "DataType",
    "DeltaTable",
    "Field",
    "Metadata",
    "PostCommitHookProperties",
    "QueryBuilder",
    "Schema",
    "TableFeatures",
    "Transaction",
    "WriterProperties",
    "__version__",
    "convert_to_deltalake",
    "rust_core_version",
    "write_deltalake",
]



#!/usr/bin/env python3
#
# This filue contains all the tests of the deltalake python package in a
# multithreaded environment

import pathlib
import threading
from concurrent.futures import ThreadPoolExecutor
from typing import TYPE_CHECKING

import pytest

from deltalake import DeltaTable, write_deltalake
from deltalake.exceptions import CommitFailedError

if TYPE_CHECKING:
    import pyarrow as pa


@pytest.mark.pyarrow
def test_concurrency(existing_table: DeltaTable, sample_data_pyarrow: "pa.Table"):
    exception = None

    def comp():
        nonlocal exception
        dt = DeltaTable(existing_table.table_uri)
        for _ in range(5):
            # We should always be able to get a consistent table state
            data = DeltaTable(dt.table_uri).to_pyarrow_table()
            # If two overwrites delete the same file and then add their own
            # concurrently, then this will fail.
            assert data.num_rows == sample_data_pyarrow.num_rows
            try:
                write_deltalake(dt.table_uri, sample_data_pyarrow, mode="overwrite")
            except Exception as e:
                exception = e

    n_threads = 2
    threads = [threading.Thread(target=comp) for _ in range(n_threads)]

    for t in threads:
        t.start()
    for t in threads:
        t.join()

    assert isinstance(exception, CommitFailedError)
    assert (
        "a concurrent transaction deleted the same data your transaction deletes"
        in str(exception)
    )


@pytest.mark.polars
def test_multithreaded_write_using_table(tmp_path: pathlib.Path):
    import polars as pl

    table = pl.DataFrame({"a": [1, 2, 3]})
    write_deltalake(tmp_path, table, mode="overwrite")

    dt = DeltaTable(tmp_path)

    with ThreadPoolExecutor() as exe:
        list(
            exe.map(
                lambda i: write_deltalake(dt, pl.DataFrame({"a": [i]}), mode="append"),
                range(5),
            )
        )


@pytest.mark.polars
def test_multithreaded_write_using_path(tmp_path: pathlib.Path):
    import polars as pl

    table = pl.DataFrame({"a": [1, 2, 3]})
    write_deltalake(tmp_path, table, mode="overwrite")

    with ThreadPoolExecutor() as exe:
        list(
            exe.map(lambda _: write_deltalake(tmp_path, table, mode="append"), range(5))
        )



from arro3.core import Table

from deltalake import write_deltalake
from deltalake.table import DeltaTable, DeltaTableConfig


def test_config_roundtrip(tmp_path, sample_table: Table):
    write_deltalake(tmp_path, sample_table)

    config = DeltaTableConfig(without_files=True, log_buffer_size=100)

    dt = DeltaTable(
        tmp_path,
        without_files=config.without_files,
        log_buffer_size=config.log_buffer_size,
    )

    assert config == dt.table_config

    config = DeltaTableConfig(without_files=False, log_buffer_size=1)

    dt = DeltaTable(
        tmp_path,
        without_files=config.without_files,
        log_buffer_size=config.log_buffer_size,
    )

    assert config == dt.table_config



from __future__ import annotations

import json
import os
from pathlib import Path
from typing import TYPE_CHECKING, Any, NamedTuple

import pytest

from deltalake import DeltaTable

if TYPE_CHECKING:
    import pyarrow as pa


class ReadCase(NamedTuple):
    root: Path
    version: int | None
    case_info: dict[str, Any]
    version_metadata: dict[str, Any]


cases = []

dat_version = "0.0.2"
reader_case_path = Path("dat-data") / f"v{dat_version}" / "reader_tests" / "generated"

if not reader_case_path.exists():
    pytest.skip(
        "DAT test data not present. Run make setup-dat to download them.",
        allow_module_level=True,
    )

for path in reader_case_path.iterdir():
    if path.is_dir():
        with open(path / "test_case_info.json") as f:
            metadata = json.load(f)

        for version_path in (path / "expected").iterdir():
            if path.name.startswith("v"):
                version = int(path.name[1:])
            else:
                version = None
            with open(version_path / "table_version_metadata.json") as f:
                version_metadata = json.load(f)

        cases.append(ReadCase(path, version, metadata, version_metadata))

failing_cases = {
    "multi_partitioned_2": "Waiting for PyArrow 11.0.0 for decimal cast support (#1078)",
    "multi_partitioned": "Test case handles binary poorly",
    "all_primitive_types": "The parquet table written with PySpark incorrectly wrote a timestamp primitive without Timezone information.",
}


@pytest.mark.pyarrow
@pytest.mark.parametrize(
    "case", cases, ids=lambda case: f"{case.case_info['name']} (version={case.version})"
)
def test_dat(case: ReadCase):
    import pyarrow.parquet as pq

    root, version, case_info, version_metadata = case

    if case_info["name"] in failing_cases:
        msg = failing_cases[case_info["name"]]
        pytest.skip(msg)

    # Get Delta Table path
    delta_root = root / "delta"

    # Load table
    dt = DeltaTable(str(delta_root), version=version)

    # Verify table paths can be found
    for path in dt.file_uris():
        assert os.path.exists(path)

    # Compare protocol versions
    assert dt.protocol().min_reader_version == version_metadata["min_reader_version"]
    assert dt.protocol().min_writer_version == version_metadata["min_writer_version"]

    # If supported protocol version, try to read, load parquet, and compare
    if dt.protocol().min_reader_version <= 1:
        version_path = "latest" if version is None else f"v{version}"
        parquet_root = root / "expected" / version_path / "table_content"
        expected = pq.read_table(parquet_root, coerce_int96_timestamp_unit="us")
        actual = dt.to_pyarrow_table()
        assert_tables_equal(expected, actual)
    else:
        # We should raise an error when attempting to read too advanced protocol
        with pytest.raises(Exception):
            dt.to_pyarrow_table()


def assert_tables_equal(first: "pa.Table", second: "pa.Table") -> None:
    import pyarrow as pa

    assert first.schema == second.schema
    sort_keys = [
        (col, "ascending")
        for i, col in enumerate(first.column_names)
        if not pa.types.is_nested(first.schema.field(i).type)
    ]
    first_sorted = first.sort_by(sort_keys)
    second_sorted = second.sort_by(sort_keys)
    assert first_sorted == second_sorted



import pathlib
from typing import TYPE_CHECKING

import pytest
from arro3.core import Table

from deltalake import CommitProperties, DeltaTable, Field, Schema, write_deltalake
from deltalake.exceptions import DeltaError
from deltalake.schema import PrimitiveType

if TYPE_CHECKING:
    pass

schema = Schema(
    fields=[
        Field("id", type=PrimitiveType("string"), nullable=True),
        Field("price", type=PrimitiveType("integer"), nullable=True),
        Field("timestamp", type=PrimitiveType("timestamp_ntz"), nullable=True),
        Field("deleted", type=PrimitiveType("boolean"), nullable=True),
    ]
)


def test_create_roundtrip_metadata(
    tmp_path: pathlib.Path,
):
    dt = DeltaTable.create(
        tmp_path,
        schema,
        name="test_name",
        description="test_desc",
        configuration={
            "delta.appendOnly": "true",
            "delta.logRetentionDuration": "interval 2 days",
        },
        commit_properties=CommitProperties(custom_metadata={"userName": "John Doe"}),
    )

    metadata = dt.metadata()

    assert metadata.name == "test_name"
    assert metadata.description == "test_desc"
    assert metadata.configuration == {
        "delta.appendOnly": "true",
        "delta.logRetentionDuration": "interval 2 days",
    }
    assert dt.history()[0]["userName"] == "John Doe"

    assert {*dt.protocol().writer_features} == {"appendOnly", "timestampNtz"}  # type: ignore


def test_create_modes(tmp_path: pathlib.Path):
    dt = DeltaTable.create(tmp_path, schema, mode="error")
    last_action = dt.history(1)[0]

    with pytest.raises(DeltaError):
        dt = DeltaTable.create(tmp_path, schema, mode="error")

    assert last_action["operation"] == "CREATE TABLE"
    with pytest.raises(DeltaError):
        dt = DeltaTable.create(tmp_path, schema, mode="append")

    dt = DeltaTable.create(tmp_path, schema, mode="ignore")
    assert dt.version() == 0

    dt = DeltaTable.create(tmp_path, schema, mode="overwrite")
    assert dt.version() == 1

    last_action = dt.history(1)[0]

    assert last_action["operation"] == "CREATE OR REPLACE TABLE"


def test_create_schema(tmp_path: pathlib.Path):
    dt = DeltaTable.create(
        tmp_path,
        schema,
    )

    assert dt.schema() == schema


@pytest.mark.skip(reason="not implemented")
def test_create_with_deletion_vectors_enabled(tmp_path: pathlib.Path):
    """append only is set to false so shouldn't be converted to a feature"""
    dt = DeltaTable.create(
        tmp_path,
        schema,
        name="test_name",
        description="test_desc",
        configuration={
            "delta.appendOnly": "false",
            "delta.enableDeletionVectors": "true",
        },
        commit_properties=CommitProperties(custom_metadata={"userName": "John Doe"}),
    )

    metadata = dt.metadata()
    protocol = dt.protocol()
    assert metadata.name == "test_name"
    assert metadata.description == "test_desc"
    assert metadata.configuration == {
        "delta.appendOnly": "false",
        "delta.enableDeletionVectors": "true",
    }
    assert protocol.min_reader_version == 3
    assert protocol.min_writer_version == 7
    assert protocol.writer_features == ["deletionVectors"]  # type: ignore
    assert protocol.reader_features == ["deletionVectors"]
    assert dt.history()[0]["userName"] == "John Doe"


def test_create_higher_protocol_versions(tmp_path: pathlib.Path):
    schema = Schema(
        fields=[
            Field("id", type=PrimitiveType("string"), nullable=True),
            Field("price", type=PrimitiveType("integer"), nullable=True),
            Field("deleted", type=PrimitiveType("boolean"), nullable=True),
        ]
    )

    dt = DeltaTable.create(
        tmp_path,
        schema,
        name="test_name",
        description="test_desc",
        configuration={
            "delta.appendOnly": "false",
            "delta.minReaderVersion": "1",
            "delta.minWriterVersion": "5",
        },
        commit_properties=CommitProperties(custom_metadata={"userName": "John Doe"}),
    )

    metadata = dt.metadata()
    protocol = dt.protocol()
    assert metadata.name == "test_name"
    assert metadata.description == "test_desc"
    assert metadata.configuration == {
        "delta.appendOnly": "false",
        "delta.minReaderVersion": "1",
        "delta.minWriterVersion": "5",
    }
    assert protocol.min_reader_version == 1
    assert protocol.min_writer_version == 5
    assert dt.history()[0]["userName"] == "John Doe"


def test_create_or_replace_existing_table(tmp_path: pathlib.Path, sample_table: Table):
    write_deltalake(table_or_uri=tmp_path, data=sample_table)
    dt = DeltaTable.create(
        tmp_path,
        schema,
        partition_by=["id"],
        mode="overwrite",
    )

    assert dt.files() == []


def test_delta_table_invalid_path_raises_error(tmp_path: pathlib.Path):
    nonexistent_path = tmp_path / "does_not_exist"
    assert not nonexistent_path.exists()

    with pytest.raises(DeltaError):
        DeltaTable(str(nonexistent_path))

    assert not nonexistent_path.exists()



from __future__ import annotations

from abc import abstractmethod
from collections.abc import Mapping
from typing import TYPE_CHECKING, Any

from deltalake._internal import DeltaFileSystemHandler, RawDeltaTable

if TYPE_CHECKING:
    import pyarrow as pa
    from pyarrow.fs import FileInfo, FileSelector


class BaseDeltaStorageHandler:
    """
    BaseDeltaStorageHandler is a concrete implementations of a PyArrow FileSystemHandler.
    """

    def __init__(
        self,
        table_uri: str,
        options: dict[str, str] | None = None,
        known_sizes: dict[str, int] | None = None,
    ) -> None:
        self._handler = DeltaFileSystemHandler(
            table_uri=table_uri, options=options, known_sizes=known_sizes
        )

    @classmethod
    def from_table(
        cls,
        table: RawDeltaTable,
        options: dict[str, str] | None = None,
        known_sizes: dict[str, int] | None = None,
    ) -> BaseDeltaStorageHandler:
        self = cls.__new__(cls)
        self._handler = DeltaFileSystemHandler.from_table(table, options, known_sizes)
        return self

    def get_type_name(self) -> str:
        return self._handler.get_type_name()

    def copy_file(self, src: str, dst: str) -> None:
        """Copy a file.

        If the destination exists and is a directory, an error is returned. Otherwise, it is replaced.
        """
        return self._handler.copy_file(src=src, dst=dst)

    def create_dir(self, path: str, recursive: bool = True) -> None:
        """Create a directory and subdirectories.

        This function succeeds if the directory already exists.
        """
        return self._handler.create_dir(path, recursive)

    def delete_dir(self, path: str) -> None:
        """Delete a directory and its contents, recursively."""
        return self._handler.delete_dir(path)

    def delete_file(self, path: str) -> None:
        """Delete a file."""
        return self._handler.delete_file(path)

    def equals(self, other: Any) -> bool:
        return self._handler.equals(other)

    def delete_dir_contents(
        self,
        path: str,
        *,
        accept_root_dir: bool = False,
        missing_dir_ok: bool = False,
    ) -> None:
        """Delete a directory's contents, recursively.

        Like delete_dir, but doesn't delete the directory itself.
        """
        return self._handler.delete_dir_contents(
            path=path,
            accept_root_dir=accept_root_dir,
            missing_dir_ok=missing_dir_ok,
        )

    def delete_root_dir_contents(self) -> None:
        """Delete the root directory contents, recursively."""
        return self._handler.delete_root_dir_contents()

    def get_file_info(self, paths: list[str]) -> list["FileInfo"]:
        """Get info for the given files.

        A non-existing or unreachable file returns a FileStat object and has a FileType of value NotFound.
        An exception indicates a truly exceptional condition (low-level I/O error, etc.).
        """
        return self._handler.get_file_info(paths)

    def move(self, src: str, dest: str) -> None:
        """Move / rename a file or directory.

        If the destination exists: - if it is a non-empty directory, an error is returned - otherwise,
        if it has the same type as the source, it is replaced - otherwise, behavior is
        unspecified (implementation-dependent).
        """
        self._handler.move_file(src=src, dest=dest)

    def normalize_path(self, path: str) -> str:
        """Normalize filesystem path."""
        return self._handler.normalize_path(path)

    @abstractmethod
    def open_input_file(self, path: str) -> "pa.PythonFile":
        """
        Open an input file for random access reading.

        Args:
            path: The source to open for reading.

        Returns:
            NativeFile
        """

    @abstractmethod
    def open_input_stream(self, path: str) -> "pa.PythonFile":
        """
        Open an input stream for sequential reading.

        Args:
            path: The source to open for reading.

        Returns:
            NativeFile
        """

    @abstractmethod
    def open_output_stream(
        self, path: str, metadata: dict[str, str] | None = None
    ) -> "pa.PythonFile":
        """
        Open an output stream for sequential writing.

        If the target already exists, existing data is truncated.

        Args:
            path: The source to open for writing.
            metadata: If not None, a mapping of string keys to string values.

        Returns:
            NativeFile
        """

    def get_file_info_selector(self, selector: "FileSelector") -> list["FileInfo"]:
        """
        Get info for the files defined by FileSelector.

        Args:
            selector: FileSelector object

        Returns:
            list of file info objects
        """
        return self._handler.get_file_info_selector(
            selector.base_dir, selector.allow_not_found, selector.recursive
        )

    def open_append_stream(self, path: str, metadata: Mapping[str, str]) -> None:
        raise NotImplementedError



"""Tests that deltalake(delta-rs) can write to tables written by PySpark."""

import os
import pathlib

import pytest
from arro3.core import Array, DataType, Table
from arro3.core import Field as ArrowField

from deltalake import DeltaTable, write_deltalake
from deltalake.exceptions import DeltaProtocolError

from .utils import assert_spark_read_equal, get_spark, run_stream_with_checkpoint


@pytest.mark.pyspark
@pytest.mark.pyarrow
@pytest.mark.integration
def test_write_basic(tmp_path: pathlib.Path):
    # Write table in Spark
    import pyarrow as pa
    import pyspark

    spark = get_spark()
    schema = pyspark.sql.types.StructType(
        [
            pyspark.sql.types.StructField(
                "c1",
                dataType=pyspark.sql.types.IntegerType(),
                nullable=True,
            )
        ]
    )
    spark.createDataFrame([(4,)], schema=schema).write.save(
        str(tmp_path),
        mode="append",
        format="delta",
    )
    # Overwrite table in deltalake
    data = pa.table({"c1": pa.array([5, 6], type=pa.int32())})

    write_deltalake(str(tmp_path), data, mode="overwrite")

    # Read table in Spark
    assert_spark_read_equal(data, str(tmp_path), sort_by="c1")


@pytest.mark.pyspark
@pytest.mark.pyarrow
@pytest.mark.integration
def test_write_invariant(tmp_path: pathlib.Path):
    # Write table in Spark with invariant
    #
    spark = get_spark()
    import delta
    import delta.pip_utils
    import delta.tables
    import pyarrow as pa
    import pyspark

    schema = pyspark.sql.types.StructType(
        [
            pyspark.sql.types.StructField(
                "c1",
                dataType=pyspark.sql.types.IntegerType(),
                nullable=True,
                metadata={
                    "delta.invariants": '{"expression": { "expression": "c1 > 3"} }'
                },
            )
        ]
    )

    delta.tables.DeltaTable.create(spark).location(str(tmp_path)).addColumns(
        schema
    ).execute()

    spark.createDataFrame([(4,)], schema=schema).write.save(
        str(tmp_path),
        mode="append",
        format="delta",
    )

    # Cannot write invalid data to the table
    invalid_data = Table(
        {
            "c1": Array(
                [6, 2],
                ArrowField("c1", type=DataType.int32(), nullable=True),
            ),
        }
    )

    with pytest.raises(
        DeltaProtocolError, match=r"Invariant \(c1 > 3\) violated by value .+2"
    ):
        # raise DeltaProtocolError("test")
        write_deltalake(str(tmp_path), invalid_data, mode="overwrite")

    # Can write valid data to the table
    valid_data = pa.table({"c1": pa.array([5, 6], type=pa.int32())})
    write_deltalake(str(tmp_path), valid_data, mode="append")

    expected = pa.table({"c1": pa.array([4, 5, 6], type=pa.int32())})
    assert_spark_read_equal(expected, str(tmp_path), sort_by="c1")


@pytest.mark.pyspark
@pytest.mark.pyarrow
@pytest.mark.integration
def test_spark_read_optimize_history(tmp_path: pathlib.Path):
    import pyarrow as pa

    ids = ["1"] * 10
    values = [10, 20, 30, 40, 50, 60, 70, 80, 90, 100]

    id_array = pa.array(ids, type=pa.string())
    value_array = pa.array(values, type=pa.int32())

    pa_table = pa.Table.from_arrays([id_array, value_array], names=["id", "value"])

    # Two writes on purpose for an optimize to occur
    write_deltalake(tmp_path, pa_table, mode="append", partition_by=["id"])
    write_deltalake(tmp_path, pa_table, mode="append", partition_by=["id"])

    dt = DeltaTable(tmp_path)
    dt.optimize.compact(partition_filters=[("id", "=", "1")])

    spark = get_spark()
    history_df = spark.sql(f"DESCRIBE HISTORY '{tmp_path}'")

    latest_operation_metrics = (
        history_df.orderBy(history_df.version.desc()).select("operationMetrics").first()
    )

    assert latest_operation_metrics["operationMetrics"] is not None


@pytest.mark.pyspark
@pytest.mark.pyarrow
@pytest.mark.integration
def test_spark_read_z_ordered_history(tmp_path: pathlib.Path):
    import pyarrow as pa

    ids = ["1"] * 10
    values = [10, 20, 30, 40, 50, 60, 70, 80, 90, 100]

    id_array = pa.array(ids, type=pa.string())
    value_array = pa.array(values, type=pa.int32())

    pa_table = pa.Table.from_arrays([id_array, value_array], names=["id", "value"])

    # Two writes on purpose for an optimize to occur
    write_deltalake(tmp_path, pa_table, mode="append", partition_by=["id"])
    write_deltalake(tmp_path, pa_table, mode="append", partition_by=["id"])

    dt = DeltaTable(tmp_path)
    dt.optimize.z_order(columns=["value"], partition_filters=[("id", "=", "1")])

    spark = get_spark()
    history_df = spark.sql(f"DESCRIBE HISTORY '{tmp_path}'")

    latest_operation_metrics = (
        history_df.orderBy(history_df.version.desc()).select("operationMetrics").first()
    )

    assert latest_operation_metrics["operationMetrics"] is not None


@pytest.mark.pyspark
@pytest.mark.pyarrow
@pytest.mark.integration
def test_spark_read_repair_run(tmp_path):
    import pyarrow as pa

    ids = ["1"] * 10
    values = [10, 20, 30, 40, 50, 60, 70, 80, 90, 100]

    id_array = pa.array(ids, type=pa.string())
    value_array = pa.array(values, type=pa.int32())

    pa_table = pa.Table.from_arrays([id_array, value_array], names=["id", "value"])

    write_deltalake(tmp_path, pa_table, mode="append")
    write_deltalake(tmp_path, pa_table, mode="append")
    dt = DeltaTable(tmp_path)
    os.remove(dt.file_uris()[0])

    dt.repair(dry_run=False)
    spark = get_spark()

    history_df = spark.sql(f"DESCRIBE HISTORY '{tmp_path}'")

    latest_operation_metrics = (
        history_df.orderBy(history_df.version.desc()).select("operationMetrics").first()
    )

    assert latest_operation_metrics["operationMetrics"] is not None


@pytest.mark.pyspark
@pytest.mark.pyarrow
@pytest.mark.integration
def test_spark_stream_schema_evolution(tmp_path: pathlib.Path):
    """https://github.com/delta-io/delta-rs/issues/3274"""
    """
    This test ensures that Spark can still read from tables following
    a schema evolution write, since old behavior was to generate a new table ID
    between schema evolution runs, which caused Spark to error thinking the table had changed.
    """
    import pyarrow as pa

    data_first_write = pa.array(
        [
            {"name": "Alice", "age": 30, "details": {"email": "alice@example.com"}},
            {"name": "Bob", "age": 25, "details": {"email": "bob@example.com"}},
        ]
    )

    data_second_write = pa.array(
        [
            {
                "name": "Charlie",
                "age": 35,
                "details": {"address": "123 Main St", "email": "charlie@example.com"},
            },
            {
                "name": "Diana",
                "age": 28,
                "details": {"address": "456 Elm St", "email": "diana@example.com"},
            },
        ]
    )

    schema_first_write = pa.schema(
        [
            ("name", pa.string()),
            ("age", pa.int64()),
            ("details", pa.struct([("email", pa.string())])),
        ]
    )

    schema_second_write = pa.schema(
        [
            ("name", pa.string()),
            ("age", pa.int64()),
            (
                "details",
                pa.struct(
                    [
                        ("address", pa.string()),
                        ("email", pa.string()),
                    ]
                ),
            ),
        ]
    )
    table_first_write = pa.Table.from_pylist(
        data_first_write, schema=schema_first_write
    )
    table_second_write = pa.Table.from_pylist(
        data_second_write, schema=schema_second_write
    )

    write_deltalake(
        tmp_path,
        table_first_write,
        mode="append",
    )

    run_stream_with_checkpoint(tmp_path.as_posix())

    write_deltalake(tmp_path, table_second_write, mode="append", schema_mode="merge")

    run_stream_with_checkpoint(tmp_path.as_posix())

    # For this test we don't care about the data, we'll just let any
    # exceptions trickle up to report as a failure
